#! /usr/local/bin/gawk -f
### NB: If you change the first line, you must also change the assignment
### of Command_line in get_command_line()!
### ====================================================================
###  @Awk-file{
###     author          = "Nelson H. F. Beebe",
###     version         = "2.03",
###     date            = "23 February 2004",
###     time            = "11:27:20 MST",
###     filename        = "epsutil.awk",
###     address         = "Center for Scientific Computing
###                        University of Utah
###                        Department of Mathematics, 322 INSCC
###                        155 S 1400 E RM 233
###                        Salt Lake City, UT 84112-0090
###                        USA",
###     telephone       = "+1 801 581 5254",
###     FAX             = "+1 801 585 1640, +1 801 581 4148",
###     URL             = "http://www.math.utah.edu/~beebe",
###     checksum        = "00426 1933 7983 58459",
###     email           = "beebe@math.utah.edu (Internet)",
###     email           = "beebe@math.utah.edu, beebe@acm.org,
###                        beebe@computer.org (Internet)",
###     codesource      = "ftp://ftp.math.utah.edu/pub/misc/index.html#epsutil-2.03",
###     codetable       = "ISO/ASCII",
###     keywords        = "background; bounding box; BoundingBox;
###                        clipping; color; EPS; Encapsulated
###                        PostScript; framing; landscape; linewidth;
###                        magnification; portrait; PostScript utility;
###                        rotation; scaling; translation",
###     supported       = "yes",
###     abstract        = "This program inserts a wrapper around an
###                        Encapsulated PostScript file to accomplish a
###                        rotation down to the right by -90 degrees,
###                        turning landscape orientation into portrait
###                        orientation.  Command-line options are
###                        available for control of background color,
###                        clipping, linewidth, scaling, and
###                        translation.",
###     docstring       = "Filter an Encapsulated PostScript file, or
###                        files, named on the command line (or stdin if
###                        no filenames are given), to stdout, inserting
###                        wrapper PostScript to modify the image.
###
###                        A detailed description can be found in the
###                        accompanying UNIX manual pages, which are
###                        available in several different formats,
###                        including HTML with pictures.
###
###                        The checksum field above contains a CRC-16
###                        checksum as the first value, followed by the
###                        equivalent of the standard UNIX wc (word
###                        count) utility output of lines, words, and
###                        characters.  This is produced by Robert
###                        Solovay's checksum utility.",
###  }
### ====================================================================

BEGIN           { pre_initialize() }

/^%!/		{ if (Pre_Scan && Pre_Init) initialize() }

/^%%EOF/        { next }        # discard EOF comments

/^%%BoundingBox: *[0-9.]+[ \t]+[0-9.]+[ \t]+[0-9.]+[ \t]+[0-9.]+/ \
		{ if (Pre_Init) initialize(); if (Pre_Scan) BoundingBox($2,$3,$4,$5) }

		{ output($0) }  # copy everything verbatim

END             { end_wrapper() }

#=======================================================================


function bad_paper_type()
{
    warning("unrecognized paper type [" PAPER "]: `" DEFAULTPAPER "' type assumed")
    PAPER = DEFAULTPAPER
}


function bigpoint(dimension)
{
    return (72.0 * inch(dimension))
}


function BoundingBox(llx,lly,urx,ury, k)
{
    LLX = floor(llx - Frame)
    LLY = floor(lly - Frame)
    URX = ceil(urx + Frame)
    URY = ceil(ury + Frame)
    start_wrapper()
}


function ceil(x)
{
    return ((int(x) < x) ? (int(x) + 1) : int(x))
}


function check_EPS_operators(s, k)
{
    k = index(s,"%")
    if (k == 0)
	k = length(s)
    if ((k > 1) && match(substr(s,1,k), \
	"banddevice|grestoreall|nulldevice|setpageparams|copypage|" \
	"initclip|quit|setsccbatch|erasepage|initgraphics|renderbands|" \
	"setscreen|exitserver|initmatrix|setdevice|settransfer|framedevice|" \
	"note|setmatrix"))
	warning("operator [" substr(s,RSTART,RLENGTH) "] is illegal in EPS file")
}


function decode_symbolic_coordinate(s)
{
    s = toupper(s)
    if (((s ~ /B/) && (s ~ /L/)) || ((s ~ /S/) && (s ~ /W/))) # BL or SW
    {
	if (s ~ /X/)
	    return (LLX)
	else if (s ~ /Y/)
	    return (LLY)
    }
    else if ((s ~ /C/) && (s ~ /L/))	# CL
    {
	if (s ~ /X/)
	    return (LLX)
	else if (s ~ /Y/)
	    return ((URY + LLY)/2)
    }
    else if (((s ~ /T/) && (s ~ /L/)) || ((s ~ /N/) && (s ~ /W/))) # TL or NW
    {
	if (s ~ /X/)
	    return (LLX)
	else if (s ~ /Y/)
	    return (URY)
    }
    else if (((s ~ /B/) && (s ~ /R/)) || ((s ~ /S/) && (s ~ /E/))) # BR or SE
    {
	if (s ~ /X/)
	    return (URX)
	else if (s ~ /Y/)
	    return (LLY)
    }
    else if ((s ~ /C/) && (s ~ /R/))	# CR
    {
	if (s ~ /X/)
	    return (URX)
	else if (s ~ /Y/)
	    return ((URY + LLY)/2)
    }
    else if (((s ~ /T/) && (s ~ /R/)) || ((s ~ /N/) && (s ~ /E/))) # TR or NE
    {
	if (s ~ /X/)
	    return (URX)
	else if (s ~ /Y/)
	    return (URY)
    }
    else if (s ~ /N/)			# N
    {
	if (s ~ /X/)
	    return ((URX + LLX)/2)
	else if (s ~ /Y/)
	    return (URY)
    }
    else if (s ~ /S/)			# S
    {
	if (s ~ /X/)
	    return ((URY + LLY)/2)
	else if (s ~ /Y/)
	    return (LLY)
    }
    else if (s ~ /W/)			# W
    {
	if (s ~ /X/)
	    return (LLX)
	else if (s ~ /Y/)
	    return ((URY + LLY)/2)
    }
    else if (s ~ /E/)			# E
    {
	if (s ~ /X/)
	    return (URX)
	else if (s ~ /Y/)
	    return ((URY + LLY)/2)
    }
    else if (s ~ /C/)			# C
    {
	if (s ~ /X/)
	    return ((URX + LLX)/2)
	else if (s ~ /Y/)
	    return ((URY + LLY)/2)
    }
    warning("cannot decode symbolic coordinate [" s "]: assuming 0")
    return (0)
}


function default_grid_x( parts)
{
    split(Paper[PAPER],parts," ")
    return (default_grid_xy(parts[1]))
}


function default_grid_xy(paper_dimension)
{
    return ("50bp,25bp," int(25*int((bigpoint(paper_dimension) - 25)/25)) \
	"bp,bp,25bp,5bp,5bp,100bp,9bp,Helvetica")
}


function default_grid_y( parts)
{
    split(Paper[PAPER],parts," ")
    return (default_grid_xy(parts[2]))
}


function define_bbox_background()
{
    print "% Step 6: set background color:  BG=" BG
    if (BG != "")
    {
        print "gsave"
	print "  ", PostScript_color(BG)
	print "  LLX LLY URX LLX sub URY LLY sub rectfill"
	print "grestore"
    }
}


function define_bbox_foreground()
{
    print "% Step 7: set foreground color: FG=" FG
    if (FG != "")
    {
	print "", PostScript_color(FG)
    }
}


function define_clipping_path()
{
    print "% Step 5: Set clipping path to ORIGINAL BoundingBox"
    print "newpath"
    print "  LLX LLY moveto"
    print "  LLX URY lineto"
    print "  URX URY lineto"
    print "  URX LLY lineto"
    print "closepath"
    print "CLIP { clip } if"
    print ""
}


function define_EPSF_wrappers()
{
    # These definitions are from p. 726 of the Adobe
    # PostScript Language Reference Manual
    print "% Include file wrapper macros from p. 726 of Adobe red book"
    print "/BeginEPSF"
    print "{"
    print "  /b4_Inc_state save def"
    print "  /dict_count countdictstack def"
    print "  /op_count count 1 sub def"
    print "  userdict begin"
    print "  /showpage { } def"
    print "  0 setgray"
    print "  0 setlinecap"
    print "  1 setlinewidth"
    print "  0 setlinejoin"
    print "  10 setmiterlimit"
    print "  [ ] 0 setdash"
    print "  newpath"
    print "  /languagelevel where"
    print "  {"
    print "    pop languagelevel"
    print "    1 ne"
    print "    {"
    print "      false setstrokeadjust"
    print "      false setoverprint"
    print "    } if"
    print "  } if"
    print "} bind def"
    print ""
    print "/EndEPSF"
    print "{"
    print "  count op_count sub {pop} repeat"
    print "  countdictstack dict_count sub {end} repeat"
    print "  b4_Inc_state restore"
    print "} bind def"
    print ""
}


function define_grid()
{
    set_grid()

    print "%%BeginResource: procset gridmacros " VERSION_NUMBER
    print "/" GridDict " 100 dict def"
    print GridDict " begin"
    print ""
    print "/FontNameX /" Gridx[10] " def"
    print "/FontNameY /" Gridy[10] " def"
    print ""
    print "/FontSizeX " PostScript_dimension(Gridx[9]) " def"
    print "/FontSizeY " PostScript_dimension(Gridy[9]) " def"
    print ""
    print "/GridLLX " PostScript_dimension(Gridx[1]) " def"
    print "/GridDX  " PostScript_dimension(Gridx[2]) " def"
    print "/GridURX " PostScript_dimension(Gridx[3]) " def"
    print ""
    print "/GridLLY " PostScript_dimension(Gridy[1]) " def"
    print "/GridDY  " PostScript_dimension(Gridy[2]) " def"
    print "/GridURY " PostScript_dimension(Gridy[3]) " def"
    print ""
    print "/GridStepX " PostScript_dimension(Gridx[5]) " def"
    print "/GridStepY " PostScript_dimension(Gridy[5]) " def"
    print ""
    print "/GridThickerFactor 5 def"
    print "/GridThickerX " PostScript_dimension(Gridx[8]) " def"
    print "/GridThickerY " PostScript_dimension(Gridy[8]) " def"
    print ""
    print "/GridUnitNameX (" Gridx[4] ") def"
    print "/GridUnitNameY (" Gridy[4] ") def"
    print ""
    print "/GridUnitSizeX 1 " Gridx[4] " def"
    print "/GridUnitSizeY 1 " Gridy[4] " def"
    print "/gx { GridUnitSizeX div cvi } def"
    print "/gy { GridUnitSizeY div cvi } def"
    print ""
    print "/TickLengthX " PostScript_dimension(Gridx[7]) " def"
    print "/TickLengthY " PostScript_dimension(Gridy[7]) " def"
    print "/TickStepX " PostScript_dimension(Gridx[6]) " def"
    print "/TickStepY " PostScript_dimension(Gridy[6]) " def"
    print ""
    print "/TickHOffset"
    print "{"
    print "  gsave"
    print "    /temp 12 string def"
    print "    FontNameY findfont FontSizeY scalefont setfont"
    print "    GridLLY round cvi temp cvs stringwidth pop"
    print "    GridURY round cvi temp cvs stringwidth pop"
    print "    max"
    print "    (xx) stringwidth pop add"
    print "  grestore"
    print "} def"
    print "/TickVOffset FontSizeX 2.4 mul def"
    print ""
    print "/Frame          % xmin dx xmax   ymin dy ymax   nminx dnx nmaxx   nminy dny nmaxy   Frame"
    print "{"
    print "    /Fnmaxy exch def"
    print "    /Fdny exch def"
    print "    /Fnminy exch def"
    print "    /Fnmaxx exch def"
    print "    /Fdnx exch def"
    print "    /Fnminx exch def"
    print "    /Fymax exch def"
    print "    /Fdy exch def"
    print "    /Fymin exch def"
    print "    /Fxmax exch def"
    print "    /Fdx exch def"
    print "    /Fxmin exch def"
    print "    /Fwidth Fxmax Fxmin sub def"
    print "    /Fheight Fymax Fymin sub def"
    print "    newpath"
    print "      Fxmin Fymin moveto"
    print "      Fxmax Fymin lineto"
    print "      Fxmax Fymax lineto"
    print "      Fxmin Fymax lineto"
    print "    closepath stroke"
    print ""
    print "    Fxmin Fymin Fwidth TickStepX TickLengthX TickX"
    print "    Fxmin Fymax Fwidth TickStepX TickLengthX neg TickX"
    print "    Fxmin Fymin Fheight TickStepY TickLengthY TickY"
    print "    Fxmax Fymin Fheight TickStepY TickLengthY neg TickY"
    print "    Fxmin Fdx Fxmax Fymax 0 0 Fnminx Fdnx Fnmaxx TickLengthX neg NumberX"
    print "    Fxmin Fdx Fxmax Fymin 0 0 Fnminx Fdnx Fnmaxx TickLengthX NumberX"
    print "    Fxmin 0 0 Fymin Fdy Fymax Fnminy Fdny Fnmaxy TickLengthY NumberY"
    print "    Fxmax 0 0 Fymin Fdy Fymax Fnminy Fdny Fnmaxy TickLengthY neg NumberY"
    print "    Fxmin GridStepX Fxmax GridThickerX"
    print "      Fymin GridStepY Fymax GridThickerY"
    print "      GridThickerFactor Grid"
    print "} def"
    print ""
    print "/Grid          % xmin dx xmax xthicker ymin dy ymax ythicker thickerfactor Grid"
    print "{"
    print "    /GThickerFactor exch def"
    print "    /GThickerY exch def"
    print "    /Gymax exch def"
    print "    /Gdy exch def"
    print "    /Gymin exch def"
    print "    /GThickerX exch def"
    print "    /Gxmax exch def"
    print "    /Gdx exch def"
    print "    /Gxmin exch def"
    print ""
    print "    /Gthickness currentlinewidth def"
    print "    gsave"
    print "      Gxmin Gdx Gxmax"
    print "      {"
    print "        dup"
    print "        newpath"
    print "          Gymin moveto"
    print "          0 Gymax Gymin sub rlineto"
    print "          GThickerX 0 gt"
    print "          {"
    print "            GThickerX round cvi mod 0 eq { GThickerFactor } { 1 } ifelse Gthickness mul setlinewidth"
    print "          }"
    print "          { pop }"
    print "          ifelse"
    print "        closepath stroke"
    print "      }"
    print "      for"
    print ""
    print "      /n 0 def"
    print "      Gymin Gdy Gymax"
    print "      {"
    print "        dup"
    print "        newpath"
    print "          Gxmin exch moveto"
    print "          Gxmax Gxmin sub 0 rlineto"
    print "          GThickerY 0 gt"
    print "          {"
    print "            GThickerY round cvi mod 0 eq { GThickerFactor } { 1 } ifelse Gthickness mul setlinewidth"
    print "          }"
    print "          { pop }"
    print "          ifelse"
    print "          /n 1 n add def"
    print "        closepath stroke"
    print "      }"
    print "      for"
    print "    grestore"
    print "} def"
    print ""
    print ""
    print "/NumberArgs         % xmin dx xmax ymin dy ymax nmin dn nmax tick NumberArgs"
    print "{"
    print "    /Ttick exch def"
    print "    /Tnmax exch def"
    print "    /Tdn exch def"
    print "    /Tnmin exch def"
    print "    /Tymax exch def         % unused"
    print "    /Tdy exch def           % unused"
    print "    /Tymin exch def"
    print "    /Txmax exch def"
    print "    /Tdx exch def"
    print "    /Txmin exch def"
    print "} def"
    print ""
    print ""
    print "/NumberX         % xmin dx xmax ymin dy ymax nmin dn nmax tick NumberX"
    print "{"
    print "    NumberArgs"
    print "    FontNameX findfont FontSizeX scalefont setfont"
    print "    /Tx Txmin def"
    print "    /temp 12 string def"
    print ""
    print "    Tnmin Tdn Tnmax"
    print "    {"
    print "        Tx Tymin moveto"
    print "        Ttick 0 lt"
    print "        { 0 FontSizeX 0.5 mul rmoveto }"
    print "        { 0 FontSizeX -1.5 mul rmoveto }"
    print "        ifelse"
    print "        round cvi temp cvs"
    print "        dup stringwidth pop -2 div 0 rmoveto"
    print "        show"
    print "        /Tx Tx Tdx add def"
    print "    }"
    print "    for"
    print "} def"
    print ""
    print "/NumberY         % xmin dx xmax ymin dy ymax nmin dn nmax tick NumberY"
    print "{"
    print "    NumberArgs"
    print "    FontNameY findfont FontSizeY scalefont setfont"
    print "    /Ty Tymin def"
    print "    /temp 12 string def"
    print "    Tnmin Tdn Tnmax"
    print "    {"
    print "        Txmin Ty moveto"
    print "        (  ) stringwidth pop Ttick 0 gt { neg } if 0 rmoveto"
    print "        round cvi temp cvs"
    print "        Ttick 0 gt { RightCenterShow } { LeftCenterShow } ifelse"
    print "        /Ty Ty Tdy add def"
    print "    }"
    print "    for"
    print "} def"
    print ""
    print "/TickX                                  % x y length dx tick TickX"
    print "{"
    print "    /Ttick exch def"
    print "    /Tdx exch def"
    print "    /Tlength exch def"
    print "    /Ty exch def"
    print "    /Tx exch def"
    print ""
    print "    Tx"
    print "    Tdx"
    print "    Tx Tlength add"
    print "    {"
    print "        dup"
    print "        newpath"
    print "        Ty moveto               % Tx already on stack"
    print "        round cvi Tdx 5 mul round cvi mod 0 eq% draw tick, double length every 5th"
    print "        { Ttick 2 mul }"
    print "        { Ttick }"
    print "        ifelse"
    print "        0"
    print "        exch"
    print "        rlineto"
    print "        closepath"
    print "        stroke"
    print "    }"
    print "    for"
    print "} def"
    print ""
    print "/TickY                                  % x y length dy tick TickY"
    print "{"
    print "    /Ttick exch def"
    print "    /Tdy exch def"
    print "    /Tlength exch def"
    print "    /Ty exch def"
    print "    /Tx exch def"
    print ""
    print "    Ty"
    print "    Tdy"
    print "    Ty Tlength add"
    print "    {"
    print "        dup"
    print "        newpath"
    print "        Tx exch moveto          % Ty already on stack"
    print "        round cvi Tdy 5 mul round cvi mod 0 eq% draw tick, double length every 5th"
    print "        { Ttick 2 mul }"
    print "        { Ttick }"
    print "        ifelse"
    print "        0"
    print "        rlineto"
    print "        closepath"
    print "        stroke"
    print "    }"
    print "    for"
    print "} def"
    print ""
    print "% (string) CenterCenterShow -- show centered at currentpoint"
    print "/CenterCenterShow"
    print "{"
    print "    dup stringwidth pop -2 div FontSizeX -2 div rmoveto show"
    print "} def"
    print ""
    print "% (string) RightCenterShow -- show right-centered at currentpoint"
    print "/RightCenterShow"
    print "{"
    print "    dup stringwidth pop neg FontSizeY -2 div rmoveto show"
    print "} def"
    print ""
    print "% (string) LeftCenterShow -- show left-centered at currentpoint"
    print "/LeftCenterShow"
    print "{"
    print "    0 FontSizeY -2 div rmoveto show"
    print "} def"
    print ""
    print "end % " GridDict
    print ""
    print "%%EndResource"
    print ""
}


function define_hooks()
{
    print "% Define default hooks"
    print "/BeginPageSetupHook { } def"
    print "/EndPageSetupHook { } def"
    print "/BeginDocumentHook { } def"
    print "/EndDocumentHook { } def"
    print "/EndJobHook { } def"
    print ""
}


function define_id( default,k,n,words)
{
    gsub(/ +/," ",ID)		# reduce multiple spaces to single ones
    n = split(ID,words,Item_sep_regexp)

    if ((n == 1) && (words[1] ~ /^[0-9]+$/) && (words[1] != 0))
	n = 0			# ID=<non-zero> means use default location, size, and font

    # supply defaults for omitted values
    default = "PageHeight 25 bp sub IDfontsize sub"
    if (n < 1) words[1] = "30bp"
    if (n < 2)
	words[2] = default
    else
	words[2] = bigpoint(words[2])
    if (n < 3) words[3] = "24bp"
    if (n < 4) words[4] = "Helvetica-Bold"

    # Because the list is long, do a sanity check on the values.  That
    # way, the user will at least be warned in case bad PostScript is
    # generated from, for example, a misspelled dimension, or
    # out-of-order font name.

    for (k = 1; k <= 4; ++k)
    {
	if (k == 4)
	{
	    if (words[k] !~ /^[A-Z][A-Za-z0-9-]+$/)
		warning("entry " k " [" words[k] \
		    "] of ID list does not match alphanumeric font name")
	}
	else if ((k == 2) && (words[k] ~ default))
	    ;
	else if (words[k] !~ \
	    /^[+\-]?[0-9]+([.][0-9]*)?(bp|cc|cm|dd|in|mm|pc|pt|sp)?$/)
	    warning("entry " k " [" words[k] \
		"] of ID list does not match dimension value")
    }

    print "%%BeginResource: procset idmacros " VERSION_NUMBER
    print "/" IDDict " 10 dict def"
    print IDDict " begin"
    print "% Define the identifying label option"
    print "/ID",((ID == "") ? "false" : "true"),"def"
    print "/IDfontsize",bigpoint(words[3]),"def"
    print "/IDfontname /" words[4],"def"
    if (IDBG != "") print "% IDBG=" IDBG
    print "/IDBG {", ((IDBG == "" ) ? ("1 1 0 setrgbcolor") : PostScript_color(IDBG)),"} def"
    if (IDFG != "") print "% IDFG=" IDFG
    print "/IDFG {", ((IDFG == "" ) ? ("0 setgray") : PostScript_color(IDFG)),"} def"
    print "/IDstring (",Command_line,") def"
    print "/IDx", bigpoint(words[1]), "def"
    print "/IDy", words[2], "def" # NB: words[2] may be token sequence
    print "end % " IDDict
    print "%%EndResource"
    print ""
}


function define_normal_state()
{
    print "/NormalState save def % remember normal state"
    print ""
}


function define_one_psfile(filename, line)
{
    print "%%BeginResource: file",filename
    while ((getline line < filename) > 0)
	print line
    close(filename)
    print "%%EndResource"
    print ""
}


function define_page_background()
{
    print "% Step 1.5: Set page background color"
    if (PAGEBG != "")
    {
        print "gsave"
	print "  % PAGEBG=" PAGEBG
	print "  ", PostScript_color(PAGEBG)
	print "  0 0 PageWidth PageHeight rectfill"
	print "grestore"
	print ""
    }
}


function define_page_dimensions( parts)
{
    split(Paper[PAPER],parts," ")
    print "% Paper type:", PAPER, " [" parts[1] " wide by " parts[2] " high]"
    print "/PageWidth ", PostScript_dimension(parts[1]), "def"
    print "/PageHeight", PostScript_dimension(parts[2]), "def"
    print "/PageMargin 25 bp def"
}


function define_page_magnification()
{
    print "% Step -1: Translate the page origin to the new page position (PTX, PTY)"
    print "PTX PTY translate"
    print ""

    print "% Step 0: Magnify the page by (MX,MY) about (PMX,PMY)"
    print "PMX PMY translate"
    print "MX MY scale"
    print "PMX neg PMY neg translate"
    print ""

    print "% Step 1: Translate the ORIGINAL figure to place its lower-left corner"
    print "%         at the new page position (OX, OY)"
    print "LLX OX sub neg LLY OY sub neg translate"
    print ""
}


function define_pscode()
{
    if (PSCODE != "")
    {
	print "% User-defined PSCODE"
        print PSCODE
	print ""
    }
}


function define_psfile( k,n,filename)
{
    if (PSFILE != "")
    {
	n = split(PSFILE,filename,"[, \t]")
	for (k = 1; k <= n; ++k)
	{
	    sub(/^[ \t]*/,"",filename[k]) # trim leading space
	    sub(/[ \t]*$/,"",filename[k]) # trim trailing space
	    define_one_psfile(filename[k])
	}
    }
}


function define_rectfill()
{
    define_rectxxx("fill")
}


function define_rectstroke()
{
    define_rectxxx("stroke")
}


function define_rectxxx(operator)
{
    print "% Define a Level 2 PostScript operator in case we are in Level 1"
    print "/rect" operator
    print "{"
    print "  /H exch def"
    print "  /W exch def"
    print "  /Y exch def"
    print "  /X exch def"
    print "  gsave"
    print "    newpath"
    print "      X Y moveto"
    print "      W 0 rlineto"
    print "      0 H rlineto"
    print "      W neg 0 rlineto"
    print "      closepath"
    print "    " operator
    print "  grestore"
    print "} def"
    print ""
}


function define_rotate_000()
{
    print "% Step 2: No rotation needed"
    print ""

    print "% Step 3: No translation needed"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL lower-left corner"
    print "LLX LLY translate"
    print "SX SY scale"
    print "LLX neg LLY neg translate"
    print ""
}


function define_rotate_090()
{
    print "% Step 2: Rotate 90 degrees up about the lower-left BoundingBox corner"
    print "LLX LLY translate"
    print "90 rotate"
    print "LLX neg LLY neg translate"
    print ""

    print "% Step 3: Translate the ORIGINAL upper-left corner to"
    print "%         the ORIGINAL lower-left corner"
    print "0 URY LLY sub neg translate"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL upper-left corner"
    print "LLX URY translate"
    print "SX SY scale"
    print "LLX neg URY neg translate"
    print ""
}


function define_rotate_180()
{
    print "% Step 2: Rotate 180 degrees up about the lower-left BoundingBox corner"
    print "LLX LLY translate"
    print "180 rotate"
    print "LLX neg LLY neg translate"
    print ""

    print "% Step 3: Translate the ORIGINAL upper-right corner to"
    print "%         the ORIGINAL lower-left corner"
    print "URX LLX sub neg URY LLY sub neg translate"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL upper-right corner"
    print "URX URY translate"
    print "SX SY scale"
    print "URX neg URY neg translate"
    print ""
}


function define_rotate_270()
{
    print "% Step 2: Rotate 90 degrees down about the lower-left BoundingBox corner"
    print "LLX LLY translate"
    print "-90 rotate"
    print "LLX neg LLY neg translate"
    print ""

    print "% Step 3: Translate the ORIGINAL lower-right corner to"
    print "%         the ORIGINAL lower-left corner"
    print "URX LLX sub neg 0 translate"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL lower-right corner"
    print "URX LLY translate"
    print "SX SY scale"
    print "URX neg LLY neg translate"
    print ""
}


function define_rotate_other(Degrees, C,S,H,theta,W)
{
    print "% Step 2: Rotate by Degrees degrees down about the lower-left BoundingBox corner"
    print "/Degrees",Degrees,"def"
    print "LLX LLY translate"
    print "Degrees rotate"
    print "LLX neg LLY neg translate"
    print ""

    # NB: These computations for the adjusted BoundingBox are only valid
    # for rotation angles small enough that no point on the right edge
    # of the original BoundingBox is positioned to the left of x=LLX after
    # the rotation.
    H = URY - LLY
    W = URX - LLX
    theta = 3.141592653589793238462643 * (Degrees / 180.0)
    C = cos(theta)
    S = sin(theta)
    if (Degrees >= 0)
    {
        print "%%RotatedBoundingBox:",\
	    floor(LLX - H*S), floor(LLY), \
	    ceil(URX + W - W*C), ceil(URY + W*S + H*C - H)
        print "%%RotatedBoundingBox:",\
	    floor(LLX - H*S), floor(LLY), \
	    ceil(URX + W - W*C), ceil(URY + W*S + H*C - H) >"/dev/stderr"
    }
    else
    {
        print "%%RotatedBoundingBox:",\
	    floor(LLX), floor(LLY + W*S), \
	    ceil(URX - H*S + W*C - W), ceil(URY - H + H*C)
        print "%%RotatedBoundingBox:",\
	    floor(LLX), floor(LLY + W*S), \
	    ceil(URX - H*S + W*C - W), ceil(URY - H + H*C) >"/dev/stderr"
    }

    print "% Step 3: No translation"
    print ""

    print "% Step 4: Scale the drawing about the ORIGINAL lower-left corner"
    print "LLX LLY translate"
    print "SX SY scale"
    print "LLX neg LLY neg translate"
    print ""
}


function define_setlinewidth()
{
    print "% Linewidth control"
    print "/MINLINEWIDTH", MinLineWidth, "def"
    print "/MAXLINEWIDTH", MaxLineWidth, "def"
    print "/LINEWIDTHSCALE", LineWidthScale, "def"
    print "/old_setlinewidth systemdict begin /setlinewidth load end def"
    print "/setlinewidth"
    print "{"
    print "  " EPSutilDict " begin"
    print "    LINEWIDTHSCALE mul"
    print "    dup MINLINEWIDTH lt { pop MINLINEWIDTH } if"
    print "    dup MAXLINEWIDTH gt { pop MAXLINEWIDTH } if"
    print "    old_setlinewidth"
    print "  end"
    print "} bind def"
    print ""
}


function define_transformation( degrees)
{
    degrees = (ROTATE == "") ? 0 : ((360 + (ROTATE % 360)) % 360)

    if (degrees == 0)
	define_rotate_000()
    else if (degrees == 90)
	define_rotate_090()
    else if (degrees == 180)
	define_rotate_180()
    else if (degrees == 270)
	define_rotate_270()
    else if ((-90 < ROTATE) && (ROTATE < 90))
	define_rotate_other(ROTATE)
    else
	warning("rotation angle must be a multiple of +/-90 degrees: saw [" ROTATE "]")
}


function define_user_procedures()
{
    print "% Define TeX coordinate conversion procedures so that all coordinate"
    print "% values can be followed by a space and a unit name to have coordinates"
    print "% in other than the default PostScript big points."
    print "/bp { } def"
    print "/cc { 12 mul dd } def"
    print "/cm { 2.54 div in } def"
    print "/dd { 1238 mul 1157 div pt } def"
    print "/in { 72 mul } def"
    print "/mm { 25.4 div in } def"
    print "/pc { 12 mul pt } def"
    print "/pt { 72 mul 72.27 div } def"
    print "/sp { 65536 div pt } def"
    print ""
    print "/max { dup 3 2 roll dup 4 1 roll exch gt {pop} {exch pop} ifelse } def"
    print ""
}


function define_user_variables()
{
    print "% Define the clipping option"
    print "/CLIP",(CLIP ? "true" : "false"),"def"
    print ""

    print "% Define BoundingBox frame linewidth (negative value suppresses)"
    print "/FRAME",((FRAME == "") ? "-1" : bigpoint(FRAME)),"def"

    print "/FRAMEFG {", ((FRAMEFG == "") ? ("0 setgray") : PostScript_color(FRAMEFG)), "} def"
    print ""

    print "% Define the ORIGINAL BoundingBox coordinates"
    print "% Units: PostScript big points: 72bp == 1in == 25.4mm"
    print "/LLX", LLX, "def"
    print "/LLY", LLY, "def"
    print "/URX", URX, "def"
    print "/URY", URY, "def"
    print ""

    print "% Define the x and y page magnification factors"
    print "/MX", MX, "def"
    print "/MY", MY, "def"
    print ""

    print "% Define the new position of the lower-left BoundingBox corner"
    print "% Units: PostScript big points: 72bp == 1in == 25.4mm"
    print "/OX", OX, "def"
    print "/OY", OY, "def"
    print ""

    print "% Define the magnfication origin"
    print "% Units: PostScript big points: 72bp == 1in == 25.4mm"
    print "/PMX", PMX, "def"
    print "/PMY", PMY, "def"
    print ""

    print "% Define the new translated position of the lower-left page corner"
    print "% Units: PostScript big points: 72bp == 1in == 25.4mm"
    print "/PTX", PTX, "def"
    print "/PTY", PTY, "def"
    print ""

    print "% Define the x and y scale factors"
    print "/SX", SX, "def"
    print "/SY", SY, "def"
    print ""
}


function do_comment_header( degrees,parts)
{
    Command_line = get_command_line()
    print "%!PS-Adobe-3.0 EPSF-3.0"
    print "%%Title:", Command_line
    print "%%CreationDate: Sat Jun 29 06:19:29 1996"
    print "%%Creator: epsutil.awk", VERSION
    print "%%For: " Personal_Name " <" UserName "@" HostName ">"

    degrees = (ROTATE == "") ? 0 : ((360 + (ROTATE % 360)) % 360)
    if ((degrees == 90) || (degrees == 270))
        print "%%BoundingBox:", floor(OX), floor(OY), \
		ceil(OX + SY*(URY - LLY)), ceil(OY + SX*(URX - LLX))
    else # (degrees == 0) || (degrees == 180)
        print "%%BoundingBox:", floor(OX), floor(OY), \
		ceil(OX + SX*(URX - LLX)), ceil(OY + SY*(URY - LLY))

    split(Paper[PAPER],parts," ")
    print "%%DocumentMedia: " PAPER " " bigpoint(parts[1]) " " bigpoint(parts[2]) " 0 () ()"

    print "%%DocumentData: Clean7Bit"
    print "%%DocumentFonts:"
    print "%%DocumentNeededResources:"
    print "%%DocumentSuppliedResources:"
    print "%%Orientation: Portrait"
    print "%%PageOrder: Ascend"
    print "%%Pages: 1"
    print "%%EndComments"
    print "%======================================================================="
}


function do_grid()
{
    print "%%BeginResource: procset gridshow " VERSION_NUMBER
    print GridDict " begin"
    print ""
    print "% Make grid lines a quarter of the normal thickness"
    print "currentlinewidth 4 div setlinewidth"
    print ""
    print "% draw the grid and number its axes"
    print "GridLLX GridDX GridURX GridLLY GridDY GridURY"
    print "GridLLX gx GridDX gx GridURX gx"
    print "GridLLY gy GridDY gy GridURY gy Frame"
    print ""
    print "FontNameX findfont FontSizeX scalefont setfont"
    print ""
    print "% label bottom horizontal axis at each end"
    print "GridLLX GridLLY moveto"
    print "0 TickVOffset neg rmoveto"
    print "GridUnitNameX CenterCenterShow"
    print ""
    print "GridURX GridLLY moveto"
    print "0 TickVOffset neg rmoveto"
    print "GridUnitNameX CenterCenterShow"
    print ""
    print "% label top horizontal axis at each end"
    print "GridLLX GridURY moveto"
    print "0 TickVOffset rmoveto"
    print "GridUnitNameX CenterCenterShow"
    print ""
    print "GridURX GridURY moveto"
    print "0 TickVOffset rmoveto"
    print "GridUnitNameX CenterCenterShow"
    print ""
    print "FontNameY findfont FontSizeY scalefont setfont"
    print ""
    print "% label left vertical axis at each end"
    print "GridLLX GridLLY moveto"
    print "TickHOffset neg 0 rmoveto"
    print "GridUnitNameY RightCenterShow"
    print "GridLLX GridURY moveto"
    print "TickHOffset neg 0 rmoveto"
    print "GridUnitNameY RightCenterShow"
    print ""
    print "% label right vertical axis at each end"
    print "GridURX GridLLY  moveto"
    print "TickHOffset 0 rmoveto"
    print "GridUnitNameY LeftCenterShow"
    print "GridURX GridURY  moveto"
    print "TickHOffset 0 rmoveto"
    print "GridUnitNameY LeftCenterShow"
    print ""
    print "end % " GridDict
    print "%%EndResource"
}


function do_header()
{
    do_comment_header()
    do_prolog()
    do_setup()
    do_page_setup()
}


function do_id( n,words)
{
    # The identifying label is drawn in a shaded box to ensure that it
    # is visible, using the systemdict version of fill (in case a
    # redefinition of fill is in effect). The requested font size is
    # automatically reduced if the string would be too long for the
    # available space.
    print "%%BeginResource: procset idshow " VERSION_NUMBER
    print IDDict " begin"
    print "ID"
    print "{"
    print "  gsave"
    print "    /IDMaxWidth PageWidth PageMargin sub PageMargin IDx max sub def"
    print "    IDfontname findfont IDfontsize scalefont setfont"
    print "    IDstring stringwidth pop"
    print "    dup IDMaxWidth gt { IDMaxWidth exch div } { pop 1 } ifelse"
    print "    IDfontsize mul /IDfontsize exch def"
    print "    IDfontname findfont IDfontsize scalefont setfont"
    print "    IDBG"
    print "    IDx IDy IDfontsize 0.5 mul sub IDstring stringwidth pop IDfontsize 1.7 mul"
    print "    systemdict begin rectfill end"
    print "    IDFG"
    print "    IDx IDy moveto"
    print "    IDstring show"
    print "  grestore"
    print "} if"
    print "end % " IDDict
    print "%%EndResource"
    print ""
}


function do_page_setup()
{
    print "%%Page: 1 1"
    print "%%BeginPageSetup"
    print EPSutilDict " begin"
    print "BeginPageSetupHook"
    define_page_magnification()	# NB: this must done BEFORE the state is saved,
				# since it applies to the entire page, including
				# any overlaid grid.
    define_page_background()
    define_normal_state()
    define_transformation()
    define_clipping_path()
    define_bbox_background()
    define_bbox_foreground()
    print "EndPageSetupHook"
    print "%%EndPageSetup"
}


function do_prolog()
{
    print "%%BeginProlog"
    print "/" EPSutilDict " 1000 dict def" # NB: large: private code may be supplied
    print EPSutilDict " begin"
    print ""
    define_hooks()
    define_user_procedures()
    define_user_variables()
    define_page_dimensions()
    define_EPSF_wrappers()
    define_setlinewidth()
    define_rectfill()
    define_rectstroke()
    if (GRID)
	define_grid()
    define_id()
    define_psfile()		# NB: handle files before command-line code,
    define_pscode()		# so the latter can override the former

    print "end % " EPSutilDict
    print "%%EndProlog"
    print "%======================================================================="
}


function do_setup()
{
    print "%%BeginSetup"
    print "/#copies 1 def"
    print "%%EndSetup"
    print "%======================================================================="
}


function do_trailer()
{
    print "EndJobHook"
    print "end % " EPSutilDict
    print "showpage"
    print "%%PageTrailer"
    print "%======================================================================="
    print "%%Trailer"
    print "%%EOF"
}


function end_wrapper()
{
    if (Pre_Scan)               # no valid BoundingBox found, so provide
	BoundingBox(0,0,612,792) # a default one for US A-size paper

    print "%%EndDocument"
    print "EndDocumentHook"
    print "EndEPSF"

    set_frame()			# must draw frame AFTER picture to ensure
				# overwriting of anything outside the
				# original BoundingBox

    print "% Return to normal graphics state"
    print "NormalState restore"
    print ""

    if (GRID)
	do_grid()

    do_id()

    do_trailer()
}


function fatal(message)
{
    warning("FATAL error: " message)
    exit(1)
}


function file_exists(filename)
{
    # Unfortunately, getline in both nawk and gawk will hang if its
    # file does not exist, so we need to test for file existence by
    # invoking a shell command, sigh...

    return (system("test -f " filename) == 0)
}


function find_file(path,filename, fullname,k)
{
    for (k = 1; k in path; ++k)
    {
	fullname = path[k] "/" filename
	if (file_exists(fullname))
	    return (fullname)
    }
    return ("")
}


function floor(x)
{
     return ((int(x) > x) ? (int(x) - 1) : int(x))
}


function get_command_line( command_line,k)
{
    command_line = "/usr/local/bin/gawk -f epsutil.awk"
    for (k = 1; k < ARGC; ++k)
    {
	if (ARGV[k] ~ /[\[\]\{\} *?"]/)
	    command_line = command_line " '" ARGV[k] "'"
	else
	    command_line = command_line " " ARGV[k]
    }
    gsub(/\n/," ",command_line)		# NB: must reduce multiple lines to single one
    return (command_line)
}


function hexval(s, k,h,v)
{
    s = tolower(s)
    h = 0
    for (k = 1; k <= length(s); ++k)
    {
	v = index("0123456789abcdef",substr(s,k,1)) - 1
	if (v < 0) # cannot happen since s is currently pattern-matched against
	{	   # [0-9a-fA-Z], but be safe in case of future changes
	    warning("bad hexadecimal digit at position " k " in [" s "]")
	    break
	}
	h = 16*h + v
    }
    return (h)
}


function in_range(a,b,c)
{
    return ((a <= b) && (b <= c))
}


function inch(dimension, number,units)
{
    number = dimension + 0	# coerce to number, ignoring units

    if (match(dimension, \
	/^[ \t]*[-+]?[0-9]+[.]?[0-9]*(bp|in|cm|pt|cc|dd|mm|pc|sp)?[ \t]*$/))
    {
	units = substr(dimension,RSTART+RLENGTH-2,2)

	if (units == "bp")	# big points
	    number /= 72.0
	else if (units == "in")	# inches
	    ; # NO-OP
	else if (units == "cm")	# centimeters
	    number /= 2.54
	else if (units == "pt")	# points
	    number /= 72.27
	else if (units == "cc")	# cicero
	    number *= 12.0 * (1238.0 / 1157.0) / 72.27
	else if (units == "dd")	# didot points
	    number *= (1238.0 / 1157.0) / 72.27
	else if (units == "mm")	# millimeters
	    number /= 25.4
	else if (units == "pc")	# pica
	    number *= 12.0 / 72.27
	else if (units == "sp")	# scaled points
	    number /= (65536.0 * 72.27)
	else			# must be dimensionless number, assume bp
	    number /= 72.0
    }
    else
	warning("unrecognized dimension [" dimension "]: bp assumed")

    return (number)
}


function initialize( word)
{
    CHECK = (CHECK == "") ? 0 : (0 + CHECK)
    CLIP  = (CLIP == "")  ? 1 : (0 + CLIP)
    GRID  = (GRID == "")  ? 0 : (0 + GRID)

    if (GRID)			# must turn off clipping when grid is used
	CLIP = 0

    Frame = (FRAME == "") ? 0 : bigpoint(FRAME)
    LineWidthScale = (LINEWIDTHSCALE == "") ? 1 : (0 + LINEWIDTHSCALE)
    MinLineWidth = bigpoint((MINLINEWIDTH == "") ? 0 : MINLINEWIDTH)
    MaxLineWidth = bigpoint((MAXLINEWIDTH == "") ? 24 : MAXLINEWIDTH)
    if (MaxLineWidth < Frame)
	MaxLineWidth = Frame
    set_paper()			# NB: MUST be called before BoundingBox()
    if (BOUNDINGBOX != "")
    {
        gsub(/ +/," ",BOUNDINGBOX) # reduce multiple spaces to single ones
	if (split(BOUNDINGBOX,word,Item_sep_regexp) == 4)
	    BoundingBox(bigpoint(word[1]),bigpoint(word[2]), \
			bigpoint(word[3]),bigpoint(word[4]))
	else
	    warning("BOUNDINGBOX option needs 4 dimensions: saw [" BOUNDINGBOX "]")
    }
    Pre_Init = 0
}


function initialize_paper_table()
{				# [Code borrowed verbatim from psposter.awk]
    # name                        width    height    left    right   bottom  top

    # ANSI (note that B folded in half becomes A, etc.) (portrait)

    Paper["A"]			= "8.5in    11in     1in     1in     1in    1in"
    Paper["B"]			= "11in     17in     1in     1in     1in    1in"
    Paper["C"]			= "17in     22in     1in     1in     1in    1in"
    Paper["D"]			= "22in     34in     1in     1in     1in    1in"
    Paper["E"]			= "34in     44in     1in     1in     1in    1in"

    # Common US designations (portrait)

    Paper["Computer-1411"]	= "14in     11in     1in     1in     1in    1in"
    Paper["Legal"]		= "8.5in    13in     1in     1in     1in    1in"
    Paper["Letter"]		= "8.5in    11in     1in     1in     1in    1in"
    Paper["US-Legal"]		= "8.5in    14in     1in     1in     1in    1in"

    # Common US designations (landscape)

    Paper["A-L"]		= "11in     8.5in    1in     1in     1in    1in"
    Paper["Computer-1411-L"]	= "11in     14in     1in     1in     1in    1in"
    Paper["Legal-L"]		= "13in     8.5in    1in     1in     1in    1in"
    Paper["Letter-L"]		= "11in     8.5in    1in     1in     1in    1in"
    Paper["US-Legal-L"]		= "14in     8.5in    1in     1in     1in    1in"

    # Additional paper and envelope sizes taken from <<Hewlett-Packard
    # LaserJet Family Paper Specification Guide>> (1992).

    Paper["COM10"]		= "4.1in    9.5in    1in     1in     1in    1in"
    Paper["DL"]			= "110mm    220mm    25mm    25mm    25mm   25mm"
    Paper["Executive"]		= "7.25in   10.5in   1in     1in     1in    1in"
    Paper["Monarch"]		= "3.9in    7.5in    1in     1in     1in    1in"

    # Miscellaneous and Imagen Ultrascript paper sizes

    Paper["A4Small"]		= "210mm    297mm    25mm    25mm    25mm   25mm"
    Paper["Ledger"]		= "11in     17in     1in     1in     1in    1in"
    Paper["LetterSmall"]	= "8.5in    11in     1in     1in     1in    1in"
    Paper["Note"]		= "8.5in    11in     1in     1in     1in    1in"

    # ISO standard paper A-series (portrait).   Note that Xn folded in half
    # becomes Xn+1 (X = A, B, C)

    Paper["A0"]			= "841mm    1189mm   25mm    25mm    25mm   25mm"
    Paper["A1"]			= "594mm    841mm    25mm    25mm    25mm   25mm"
    Paper["A2"]			= "420mm    594mm    25mm    25mm    25mm   25mm"
    Paper["A3"]			= "297mm    420mm    25mm    25mm    25mm   25mm"
    Paper["A4"]			= "210mm    297mm    25mm    25mm    25mm   25mm"
    Paper["A5"]			= "148mm    210mm    25mm    25mm    25mm   25mm"
    Paper["A6"]			= "105mm    148mm    25mm    25mm    25mm   25mm"
    Paper["A7"]			= "74mm     105mm    25mm    25mm    25mm   25mm"
    Paper["A8"]			= "52mm     74mm     25mm    25mm    25mm   25mm"
    Paper["A9"]			= "37mm     52mm     25mm    25mm    25mm   25mm"
    Paper["A10"]		= "26mm     37mm     25mm    25mm    25mm   25mm"

    # ISO standard paper A-series (landscape).  Note that the final L
    # is NOT prefixed by a hyphen.

    Paper["A0L"]		= "1189mm   841mm    25mm    25mm    25mm   25mm"
    Paper["A1L"]		= "841mm    594mm    25mm    25mm    25mm   25mm"
    Paper["A2L"]		= "594mm    420mm    25mm    25mm    25mm   25mm"
    Paper["A3L"]		= "420mm    297mm    25mm    25mm    25mm   25mm"
    Paper["A4L"]		= "297mm    210mm    25mm    25mm    25mm   25mm"
    Paper["A5L"]		= "210mm    148mm    25mm    25mm    25mm   25mm"
    Paper["A6L"]		= "148mm    105mm    25mm    25mm    25mm   25mm"
    Paper["A7L"]		= "105mm    74mm     25mm    25mm    25mm   25mm"
    Paper["A8L"]		= "74mm     52mm     25mm    25mm    25mm   25mm"
    Paper["A9L"]		= "52mm     37mm     25mm    25mm    25mm   25mm"
    Paper["A10L"]		= "37mm     26mm     25mm    25mm    25mm   25mm"

    # ISO intermediate paper B-series (portrait)

    Paper["B0"]			= "1000mm   1414mm   25mm    25mm    25mm   25mm"
    Paper["B1"]			= "707mm    1000mm   25mm    25mm    25mm   25mm"
    Paper["B2"]			= "500mm    707mm    25mm    25mm    25mm   25mm"
    Paper["B3"]			= "353mm    500mm    25mm    25mm    25mm   25mm"
    Paper["B4"]			= "250mm    353mm    25mm    25mm    25mm   25mm"
    Paper["B5"]			= "176mm    250mm    25mm    25mm    25mm   25mm"
    Paper["B6"]			= "125mm    176mm    25mm    25mm    25mm   25mm"

    # ISO intermediate paper B-series (landscape)

    Paper["B0L"]		= "1414mm   1000mm   25mm    25mm    25mm   25mm"
    Paper["B1L"]		= "1000mm   707mm    25mm    25mm    25mm   25mm"
    Paper["B2L"]		= "707mm    500mm    25mm    25mm    25mm   25mm"
    Paper["B3L"]		= "500mm    353mm    25mm    25mm    25mm   25mm"
    Paper["B4L"]		= "353mm    250mm    25mm    25mm    25mm   25mm"
    Paper["B5L"]		= "250mm    176mm    25mm    25mm    25mm   25mm"
    Paper["B6L"]		= "176mm    125mm    25mm    25mm    25mm   25mm"

    # ISO envelopes for A-series

    Paper["C0"]			= "1294mm   916mm    25mm    25mm    25mm   25mm"
    Paper["C1"]			= "916mm    647mm    25mm    25mm    25mm   25mm"
    Paper["C2"]			= "647mm    458mm    25mm    25mm    25mm   25mm"
    Paper["C3"]			= "458mm    323mm    25mm    25mm    25mm   25mm"
    Paper["C4"]			= "323mm    229mm    25mm    25mm    25mm   25mm"
    Paper["C5"]			= "229mm    161mm    25mm    25mm    25mm   25mm"
    Paper["C6"]			= "161mm    114mm    25mm    25mm    25mm   25mm"

    # British paper sizes (some are also used in the USA)

    Paper["Octavo"]		= "5in      8in      1in     1in     1in    1in"
    Paper["Sixmo"]		= "6.5in    8in      1in     1in     1in    1in"
    Paper["Quarto"]		= "8in      10in     1in     1in     1in    1in"
    Paper["Foolscap"]		= "8.5in    13in     1in     1in     1in    1in"
    Paper["Government-legal"]	= "8.5in    13in     1in     1in     1in    1in"
    Paper["Folio"]		= "8.3in    13in     1in     1in     1in    1in"

    # Additional paper sizes known to ghostscript (a few of its paper sizes conflict
    # with settings above, and so are ignored):
    Paper["ArchA"]		= "9in      12in     1in     1in     1in    1in"
    Paper["ArchB"]		= "12in     18in     1in     1in     1in    1in"
    Paper["ArchC"]		= "18in     24in     1in     1in     1in    1in"
    Paper["ArchD"]		= "24in     36in     1in     1in     1in    1in"
    Paper["ArchE"]		= "36in     48in     1in     1in     1in    1in"
    Paper["Flsa"]		= "8.5in    13in     1in     1in     1in    1in"
    Paper["Flsb"]		= "8.5in    13in     1in     1in     1in    1in"
    Paper["HalfLetter"]		= "5.5in    8.5in    1in     1in     1in    1in"
}


function output(s)
{
    if (CHECK)
	check_EPS_operators(s)
    if (Pre_Scan)
	Lines[++N_Lines] = s
    else
	print s
}


function PostScript_color(value, k,n,words)
{
    # Decode a general color specification, and return a PostScript
    # command to set the color.  The value may be in any of the supported
    # color models.
    n = split(value,words,Item_sep_regexp)
    if (value ~ /^\#[0-9a-fA-Z]+$/)
	return (set_rgb(value))		# X hex color
    else if ((n == 1) && ((COLORMODEL == "X") || (words[1] ~ /^[A-Za-z][A-Za-z0-9]*/)))
	return (set_named_color(value))	# looks like named color
    else if ((n == 3) && (COLORMODEL == "HSB"))
	return (set_hsb(value))
    else if ((n == 3) && (COLORMODEL == "RGB"))
	return (set_rgb(value))
    else if ((n == 4) && (COLORMODEL == "CMYK"))
	return (set_cmyk(value))
    else if ((n == 1) && (COLORMODEL == "GRAY"))
	return (set_gray(value))
    else
    {
	warning("color value [" value "] does not conform to color model [" COLORMODEL "]")
	return ("% Bad color value [" value "]")
    }
}


function PostScript_dimension(s)
{
    # convert "2.54cm" to "2.54 cm"
    gsub(/[ ]/,"",s)		# eliminate internal spaces
    if (s ~ /(bp|in|cm|pt|cc|dd|mm|pc|sp)$/)
	return ((substr(s,1,length(s)-2) " " substr(s,length(s)-1)))
    else if (s ~ /[.0-9]$/)
	return (s " bp")
    else
    {
	warning("bad dimension [" s "]: 0 bp assumed")
	return ("0 bp")
    }
}


function pre_initialize( cmd,date,pid)
{
    # Change these two lines whenever the program is modified
    VERSION_NUMBER = "2.03"
    VERSION = "Version " VERSION_NUMBER " [23-Feb-2004]"

    # Create unique dictionary names by time stamping and, when
    # available, PID stamping, so that multiple passes through epsutil
    # do not create confusion about which dictionary a particular
    # variable comes from.  PID stamps will disambiguate the case of
    # this program being run multiple times in a pipeline, whereas time
    # stamping may not.  Input from "/dev/pid" returns an empty string
    # with awk, nawk, and mawk; however, with mawk, it also generates
    # a warning "/bin/sh: :/: not found:.
    #
    # At epsutil version 2.01, it appears that the only variables in
    # need of dictionary specification are the three (LINEWIDTHSCALE,
    # MAXLINEWIDTH, and MINLINEWIDTH) referenced inside the setlinewidth
    # redefinition.  All others appear to be used with the correct
    # dictionary at the top of the dictionary stack.  Nevertheless, for
    # future use, we make unique names for each dictionary.

    if (("pid" in PROCINFO) && ((0 + PROCINFO["pid"]) > 0))
	pid = PROCINFO["pid"]
    else if ((getline pid < "/dev/pid") > 0) # use gawk extension when available
	;
    if (pid != "")
	pid = "_" pid

    "date" | getline date
    close("date")

    gsub(/ +/,"_",date)
    EPSutilDict = "EPSutilDict_" date pid
    IDDict = "IDDict_" date pid
    GridDict = "GridDict_" date pid

    "date" | getline Current_Date_and_Time
    close("date")

    if ("USER" in ENVIRON)
	UserName = ENVIRON["USER"]
    else if ("LOGNAME" in ENVIRON)
	UserName = ENVIRON["LOGNAME"]
    else
	UserName = "unknown"
    cmd = "awk -F: '/^" UserName ":/ {print $5}' /etc/passwd"
    cmd | getline Personal_Name
    close(cmd)

    if (Personal_Name == "")
    {
	cmd = "ypcat passwd | awk -F: '/^" UserName ":/ {print $5}'"
        cmd | getline Personal_Name
	close(cmd)
    }

    "hostname" | getline HostName
    close("hostname")

    N_Lines = 0
    Pre_Scan = 1
    Pre_Init = 1

    if (COLORMODEL == "") COLORMODEL = "RGB"
    initialize_paper_table()

    # CUSTOMIZATION: Choose one of these defaults:
    DEFAULTPAPER = "A"		# default is American A-format (8.5in x 11in) paper
    # DEFAULTPAPER = "A4"		# default is ISO A4-format (210mm x 297mm) paper

    Item_sep_regexp = "[^-0-9.A-Za-z+]"
}


function set_BoundingBox_origin()
{
    OX = bigpoint((OX != "") ? OX : ((OXY != "") ? OXY : LLX))
    OY = bigpoint((OY != "") ? OY : ((OXY != "") ? OXY : LLY))
}


function set_cmyk(cmyk)
{
    return (set_color("CMYK",cmyk,4,"setcmykcolor"))
}


function set_color(name,value,count,operator, words)
{
    if ((split(value,words,Item_sep_regexp) == count) && \
	in_range(0,words[1],1) && \
	in_range(0,words[2],1) && \
	in_range(0,words[3],1) && \
	in_range(0,words[count],1))
    {
	if (count == 1)
	    return (words[1] " " operator)
	else if (count == 3)
	    return (words[1] " " words[2] " " words[3] " " operator)
	else if (count == 4)
	    return (words[1] " " words[2] " " words[3] " " words[4] " " operator)
	else
	{
	    warning("internal error: bad count [" count "] passed to set_color()")
	    return ("")
	}
    }
    else
    {
	warning(name " variable needs " count " number(s) in 0..1; saw [" value "]")
	return ("")
    }
}


function set_frame()
{
    print "% Draw optional BoundingBox frame"
    print "FRAME 0 ge"
    print "{"
    print "  gsave"
    print "    0 setgray"
    print "    FRAME setlinewidth"
    print "    FRAMEFG"
    print "    LLX 0.5 FRAME mul add LLY 0.5 FRAME mul add"
    print "    URX LLX sub FRAME sub URY LLY sub FRAME sub rectstroke"
    print "  grestore"
    print "} if"
    print ""
}


function set_gray(gray)
{
    return (set_color("GRAY",gray,1,"setgray"))
}


function set_grid( s)
{
    s = default_grid_x()
    set_one_grid(((GRIDX == "") ? s : GRIDX), s, Gridx)
    s = default_grid_y()
    set_one_grid(((GRIDY == "") ? s : GRIDY), s, Gridy)
}


function set_hsb(hsb)
{
    return (set_color("HSB",hsb,3,"sethsbcolor"))
}


function set_named_color(color_value, command,grep_command,rgb,rgbfile,rgbline,rgbpath)
{
    if (RGBFILE == "")
    {
	RGBPATH = ENVIRON["RGBPATH"]
	if (RGBPATH == "")		# supply default fallback search path
	    RGBPATH = "/usr/lib/X11:/usr/openwin/lib/X11:/usr/local/lib/X11:/usr/local/X11R5/lib:/usr/lpp/X11/lib/X11"
	split(RGBPATH,rgbpath,":")
	rgbfile = find_file(rgbpath,"rgb.txt")
    }
    else
	rgbfile = RGBFILE

    command = ""
    if (rgbfile == "")
	warning("Cannot find X11 color mapping file [rgb.txt]")
    else
    {
	grep_command = ("grep 2>/dev/null -i '^[0-9 \t]*" color_value "[ \t]*$' " rgbfile)
	grep_command | getline rgbline
	close (grep_command)
	if (split(rgbline,rgb," ") > 3)
	    command = set_color("RGB", \
	        (rgb[1] / 255) " " (rgb[2] / 255) " " (rgb[3] / 255), \
	        3, "setrgbcolor")
	else
	    warning("named color [" color_value "] not found in " rgbfile)
    }
    return (command)
}


function set_one_grid(grid,default,gridxy, parts,k,nd,ng,s)
{
    # Separate the grid parameters, with omitted ones supplied by
    # default values determined by the paper type.

    gsub(/ +/," ",default)	# reduce multiple spaces to single ones
    nd = split(default, parts,  Item_sep_regexp)

    gsub(/ +/," ",grid)		# reduce multiple spaces to single ones
    ng = split(grid,    gridxy, Item_sep_regexp)

    for (k = 1; k <= ((nd > ng) ? nd : ng); ++k)
    {
	if (gridxy[k] == "")
	    gridxy[k] = parts[k]
    }

    ## printf("%% DEBUG: grid")
    ## for (k = 1; k <= 10; ++k)
    ##    printf(" [%d]=%s",k,gridxy[k])
    ## print ""

    # Because the list is long, do a sanity check on the values.  That
    # way, the user will at least be warned in case bad PostScript is
    # generated from, for example, a misspelled dimension, or
    # out-of-order font name.

    for (k = 1; k <= 10; ++k)
    {
	if (k == 4)
	{
	    if (gridxy[k] !~ /^(bp|cc|cm|dd|in|mm|pc|pt|sp)$/)
		fatal("entry " k " [" gridxy[k] \
		    "] GRID[XY] list does not match dimension name")
	}
	else if (k == 10)
	{
	    if (gridxy[k] !~ /^[A-Z][A-Za-z0-9\-]+$/)
		fatal("entry " k " [" gridxy[k] \
		    "] of GRID[XY] list does not match alphanumeric font name")
	}
	else if (gridxy[k] !~ \
	    /^[+\-]?[0-9]+([.][0-9]*)?(bp|cc|cm|dd|in|mm|pc|pt|sp)?$/)
	    fatal("entry " k " [" gridxy[k] \
		"] of GRID[XY] list does not match dimension value")
    }
    if (bigpoint(gridxy[1]) < bigpoint(gridxy[3]))
    {
	if (bigpoint(gridxy[2]) <= 0)
	    fatal("Non-positive grid step " gridxy[2])
	else if ((bigpoint(gridxy[3]) - bigpoint(gridxy[1]))/ \
		  bigpoint(gridxy[2]) > 100)
	    fatal("Too many axis numbers requested: from " gridxy[1] \
		" to " gridxy[3] " in steps of " gridxy[2])
    }
    else
	fatal("Non-positive grid interval " gridxy[1] " .. " gridxy[3])
    if (bigpoint(gridxy[5]) <= 1)
	fatal("Grid step too small " gridxy[5])
}


function set_page_magnification()
{
    MX = (MX != 0) ? MX : ((MXY != 0) ? MXY : 1)
    MY = (MY != 0) ? MY : ((MXY != 0) ? MXY : 1)
}


function set_page_magnification_origin()
{
    if (PMXY ~ /^[Cc]$/)	# handle special case of BoundingBox center
    {
	PMX = "CX"
	PMY = "CY"
        PMXY = ""
    }
    if (PMXY ~ /^[A-Za-z]+$/) PMXY = decode_symbolic_coordinate(PMXY)
    if (PMX ~ /^[A-Za-z]+$/)  PMX  = decode_symbolic_coordinate(PMX)
    if (PMY ~ /^[A-Za-z]+$/)  PMY  = decode_symbolic_coordinate(PMY)
    PMX = bigpoint((PMX != "") ? PMX : ((PMXY != "") ? PMXY : 0))
    PMY = bigpoint((PMY != "") ? PMY : ((PMXY != "") ? PMXY : 0))
}


function set_page_translation_origin()
{
    PTX = bigpoint((PTX != "") ? PTX : ((PTXY != "") ? PTXY : 0))
    PTY = bigpoint((PTY != "") ? PTY : ((PTXY != "") ? PTXY : 0))
}


function set_paper( name,parts)
{
    if (PAPER == "")		# no paper type given
	PAPER = DEFAULTPAPER
    else if (!(PAPER in Paper))	# unrecognized type
    {
        gsub(/ +/," ",PAPER)	# reduce multiple spaces to single ones
        if (split(PAPER,parts,Item_sep_regexp) == 2) # expect "width,height"
	    Paper[PAPER] = parts[1] " " parts[2] " 25mm 25mm 25mm 25mm"
	else
	{
	    warning("unrecognized paper type [" PAPER "]: A type assumed")
	    PAPER = "A"
	    warning("paper types recognized:")
	    for (name in Paper)
	    {
	        split(Paper[name],parts," ")
		printf("%-16s %7s %7s\n", name, parts[1], parts[2]) | "sort >&2"
	    }
	}
	if (!(PAPER in Paper))	# perform a final sanity check
	{
	    warning("FATAL error: cannot find default " PAPER "-type paper specifications")
	    exit(1)
	}
    }
}


function set_rgb(rgb, r,g,b)
{
    if (rgb ~ /^\#[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]$/)
    {
	r = hexval(substr(rgb,2,1))/15.0
	g = hexval(substr(rgb,3,1))/15.0
	b = hexval(substr(rgb,4,1))/15.0
	rgb = r " " g " " b
    }
    else if (rgb ~ /^\#[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]$/)
    {
	r = hexval(substr(rgb,2,2))/255.0
	g = hexval(substr(rgb,4,2))/255.0
	b = hexval(substr(rgb,6,2))/255.0
	rgb = r " " g " " b
    }
    else if (rgb ~ /^\#[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]$/)
    {
	r = hexval(substr(rgb,2,3))/4095.0
	g = hexval(substr(rgb,5,3))/4095.0
	b = hexval(substr(rgb,8,3))/4095.0
	rgb = r " " g " " b
    }
    else if (rgb ~ /^\#[0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]$/)
    {
	r = hexval(substr(rgb, 2,4))/65535.0
	g = hexval(substr(rgb, 6,4))/65535.0
	b = hexval(substr(rgb,10,4))/65535.0
	rgb = r " " g " " b
    }
    return (set_color("RGB",rgb,3,"setrgbcolor"))
}


function set_scale()
{
    if ((HEIGHT != "") || (WIDTH != "")) # absolute scaling
    {
	if (HEIGHT == "")	# uniform scaling
	    SX = SY = bigpoint(WIDTH)  / (URY - LLY)
	else if (WIDTH == "")	# uniform scaling
	    SX = SY = bigpoint(HEIGHT) / (URX - LLX)
	else			# nonuniform scaling
	{
	    SX = bigpoint(HEIGHT) / (URX - LLX)
	    SY = bigpoint(WIDTH)  / (URY - LLY)
	}
    }
    else			# relative scaling
    {
	SX = (SX != 0) ? SX : ((SXY != 0) ? SXY : 1)
	SY = (SY != 0) ? SY : ((SXY != 0) ? SXY : 1)
    }
}


function start_wrapper( k)
{
    set_page_translation_origin()
    set_page_magnification_origin()
    set_page_magnification()
    set_BoundingBox_origin()
    set_scale()
    do_header()

    print "BeginEPSF"
    print "BeginDocumentHook"
    print "%%BeginDocument:",((FILENAME == "") ? "stdin" : FILENAME)

    for (k = 1; k <= N_Lines; ++k) # print saved lines
	print Lines[k]

    Pre_Scan = 0 # done with prescan: output() will now copy instead of save
}


function warning(message)
{
    print FILENAME ":" FNR ":%%" message >"/dev/stderr"
    print "%%WARNING: " FILENAME ":" FNR ":%%" message
}
